/**
 * 
 */
package weka.classifiers.timeseries;

import java.io.PrintStream;
import java.io.Serializable;
import java.util.ArrayList;
import java.util.List;
import java.util.Collections;

import weka.classifiers.evaluation.NumericPrediction;
import weka.classifiers.timeseries.eval.ErrorModule;
import weka.core.Instance;
import weka.core.Instances;
import weka.core.Utils;

/**
 * @author mhall
 *
 */
public class ErrorBasedConfidenceLimitEstimator implements Serializable {
  
  /**
   * 
   */
  private static final long serialVersionUID = -2748314799535071043L;

  /**
   * Holds confidence bound offsets for targets at a certain level. 
   * The outer list corresponds to the fields to forecast (in the same
   * order as supplied to the TSForecaster.setFieldsToForecast() method. The inner
   * lists hold two element arrays of doubles (upper, lower bounds). The
   * first index of these lists are bounds for to a one-step-ahead forecast,
   * the second a two-step-ahead forecast, and so on 
   */
  protected List<List<double[]>> m_confidenceLimitsForTargets;
  
  /** The names of the target field(s) */
  protected List<String> m_targetFields;
  
  /** The confidence level for the limits */
  protected double m_confidenceLevel;
  
  /**
   * Get the confidence bound offsets for each target at the supplied
   * confidence level
   * 
   * @param confidenceLevel the confidence level to use
   * @return a List of confidence offsets - one for each target. Each
   * target's confidence limits are stored in a two element array,
   * where the first element stores the upper bound and the second
   * the lower bound (both are expressed as an offset)
   * @throws Exception if the confidence limits can't be computed
   * for some reason
   */
  public List<double[]> getConfidenceOffsets(double confidenceLevel,
      List<List<NumericPrediction>> predictions)
    throws Exception {
    
    if (predictions == null || predictions.get(0).size() == 0) {
      throw new Exception("No predictions have been seen yet!");
    }
    
    List<double[]> result = new ArrayList<double[]>();
    
    for (int i = 0; i < m_targetFields.size(); i++) {
      List<NumericPrediction> preds = predictions.get(i);
      
      // need to separate the positive and negative errors
      // into two separate lists
      List<Double> posErrs = new ArrayList<Double>();
      List<Double> negErrs = new ArrayList<Double>();
      
      for (NumericPrediction p : preds) {
        if (!Utils.isMissingValue(p.error())) {
          if (p.error() < 0) {
            negErrs.add(new Double(Math.abs(p.error())));
          }
          
          if (p.error() > 0) {
            posErrs.add(new Double(p.error()));
          }
        }
      }
      
      // sort into ascending order
      Collections.sort(posErrs);
      Collections.sort(negErrs);
      double[] bounds = new double[2];
      bounds[0] = Utils.missingValue();
      bounds[1] = Utils.missingValue();
      
      if (posErrs.size() > 0 && negErrs.size() > 0) {
        double cL = 1.0 - confidenceLevel;
        int posPosition = (int)Math.round(posErrs.size() * cL);
        if (posPosition < 1) {
          posPosition = 1;
        }
        int negPosition = (int)Math.round(negErrs.size() * cL);
        if (negPosition < 1) {
          negPosition = 1;
        }

        double upperBound = posErrs.get(posErrs.size() - posPosition);
        double lowerBound = negErrs.get(negErrs.size() - negPosition);
        lowerBound = -lowerBound;
        bounds[0] = lowerBound;
        bounds[1] = upperBound;
        /*System.err.println("Neg pos " + negPosition);
        System.err.println("lower: " + lowerBound + " Upper " + upperBound);*/
      }
      
      result.add(bounds);
    }

    return result;
  }
  
  public void calculateConfidenceOffsets(TSForecaster forecaster,
      Instances insts, int numPrime, int numSteps, double confidenceLevel, 
      PrintStream... progress) throws Exception {
    calculateConfidenceOffsets(forecaster, insts, numPrime, -1, numSteps, 
        confidenceLevel, progress);    
  }

  // artificialTimeStartValue is assumed to be the time value for
  // the first instance in the supplied set of instances
  public void calculateConfidenceOffsets(TSForecaster forecaster,
      Instances insts, int numPrime, int artificialTimeStartValue, 
      int numSteps, double confidenceLevel, PrintStream... progress)
      throws Exception {
    
    if (insts.numInstances() < (numPrime + numSteps)) {
      throw new Exception("We need at least " + (numPrime + numSteps) 
          + " instances in order to calculate confidence limits!");
    }
    
    if (confidenceLevel < 0 || confidenceLevel > 1) {
      throw new Exception("Confidence level must lie between 0 and 1");
    }
    
    m_targetFields = 
      AbstractForecaster.stringToList(forecaster.getFieldsToForecast());
    m_confidenceLevel = confidenceLevel;
    
    List<ErrorModule> confidenceCalculators = new ArrayList<ErrorModule>();
    for (int i = 0; i < numSteps; i++) {
      ErrorModule m = new ErrorModule();
      m.setTargetFields(m_targetFields);
      confidenceCalculators.add(m);
    }
    
    Instances primeInsts = new Instances(insts, 0, numPrime);
/*    for (int i = 0; i < numPrime; i++) {
      primeInsts.add(insts.instance(i));
    } */
    primeInsts.compactify();
    if (forecaster instanceof TSLagUser && artificialTimeStartValue >= 0) {
      ((TSLagUser)forecaster).getTSLagMaker().
        setArtificialTimeStartValue(artificialTimeStartValue - 1 + numPrime);
    }    
    
    for (int i = numPrime; i < insts.numInstances(); i++) {
      forecaster.primeForecaster(primeInsts);
      
      if (i % 10 == 0) {
        for (PrintStream p : progress) {
          p.println("Computing confidence intervals: processed " + i + " instances...");
        }
      }
      
      List<List<NumericPrediction>> forecastForSteps = 
        forecaster.forecast(numSteps);
      
      // single target only at present
      //List<NumericPrediction> preds = forecastForTargets.get(0);
      
      // update the error modules
      for (int j = 0; j < numSteps && 
        (i + j < insts.numInstances()); j++) {
        
        Instance toPredict = insts.instance(i + j);
//        double[] forecastsForStepJ = new double[m_targetFields.size()];
        List<NumericPrediction> predsForTargets = forecastForSteps.get(j);
        
/*        for (int k = 0; k < m_targetFields.size(); k++) {
          forecastsForStepJ[k] = predsForTargets.get(k).predicted();
        } */
        
        confidenceCalculators.get(j).evaluateForInstance(predsForTargets, toPredict);
      }
      
      // remove the first instance from the primeInsts and then add instance i
      // to the end
      primeInsts.delete(0);
      primeInsts.add(insts.instance(i));
      primeInsts.compactify();      
    }
    
    m_confidenceLimitsForTargets = new ArrayList<List<double[]>>();
    for (int j = 0; j < m_targetFields.size(); j++) {
      ArrayList<double[]> limitsForSingleTarget = new ArrayList<double[]>();
      for (int i = 0; i < numSteps; i++) {
        List<List<NumericPrediction>> predsForStepI = 
          confidenceCalculators.get(i).getPredictionsForAllTargets();
        List<double[]> confOffsetsForStepI = 
          getConfidenceOffsets(confidenceLevel, predsForStepI);
        
        double[] limitsAtStepI = confOffsetsForStepI.get(j);
        limitsForSingleTarget.add(limitsAtStepI);
      }
      m_confidenceLimitsForTargets.add(limitsForSingleTarget);
    }    
  }

  /* (non-Javadoc)
   * @see weka.classifiers.timeseries.ConfidenceLimitEstimator#getConfidenceLevel()
   */
  public double getConfidenceLevel() {

    return m_confidenceLevel;
  }

  /* (non-Javadoc)
   * @see weka.classifiers.timeseries.ConfidenceLimitEstimator#getConfidenceLimitsForTarget(java.lang.String)
   */
  public double[] getConfidenceLimitsForTarget(String targetName, 
      double targetValue, int stepNum)
      throws Exception {

    int index = m_targetFields.indexOf(targetName);
    
    if (index < 0) {
      throw new Exception("[ErrorBasedConfidenceLimitEstimator] " +
      		"unknown target: " + targetName);
    }
    List<double[]> confForTarget = m_confidenceLimitsForTargets.get(index); 
    if (stepNum > confForTarget.size()) {
      throw new Exception("[ErrorBasedConfidenceLimitEstimator] no limits availalbe for" +
      		"requested step number: " + stepNum);
    }
    
    double[] offsets = confForTarget.get(stepNum - 1);
    double[] limits = new double[2];
    limits[0] = targetValue + offsets[0];
    limits[1] = targetValue + offsets[1];
    
    //return confForTarget.get(stepNum - 1);
    return limits;
  }
}
