package weka.classifiers.timeseries;

import java.io.PrintStream;
import java.io.Serializable;
import java.util.ArrayList;
import java.util.Collections;
import java.util.Enumeration;
import java.util.List;
import java.util.Vector;

import weka.classifiers.AbstractClassifier;
import weka.classifiers.Classifier;
import weka.classifiers.evaluation.NumericPrediction;
import weka.classifiers.functions.SMOreg;
import weka.core.DenseInstance;
import weka.core.Instance;
import weka.core.Instances;
import weka.core.Option;
import weka.core.OptionHandler;
import weka.core.Utils;
import weka.filters.Filter;
import weka.filters.unsupervised.attribute.Remove;
import weka.filters.unsupervised.attribute.RemoveType;

public class WekaForecaster extends AbstractForecaster 
  implements TSLagUser, OptionHandler, Serializable {
  
  /** For serialization */
  private static final long serialVersionUID = 5562710925011828590L;

  /** The format of the original incoming instances */
  protected Instances m_originalHeader;
  
  /** A copy of the input data provided to primeForecaster() */
  protected transient Instances m_primedInput;
  
  /** The format of the transformed data */
  protected Instances m_transformedHeader;

  /** The base regression scheme to use */
  protected Classifier m_forecaster = new SMOreg();
  
  /** The individual forecasters for each target */
  protected List<SingleTargetForecaster> m_singleTargetForecasters;
  
  /** True if the forecaster has been built */
  protected boolean m_modelBuilt = false;          

  /** True if an artificial time index has been added to the data */
  protected boolean m_useArtificialTimeIndex = false;    
  
  /**
   * The estimator used for calculating confidence limits.
   */
  protected ErrorBasedConfidenceLimitEstimator m_confidenceLimitEstimator;
  
  /**
   * Number of steps ahead to calculate confidence limits for (0 = don't
   * calculate confidence limits 
   */
  protected int m_calculateConfLimitsSteps = 0;
  
  /** Confidence level to compute confidence limits at */
  protected double m_confidenceLevel = 0.95;
  
  /** The lag maker to use */
  TSLagMaker m_lagMaker = new TSLagMaker();
  
  /** For removing any date attributes (TSLagMaker will remap date timestamps to numeric) */
  protected RemoveType m_dateRemover;
  
  /** 
   * Holds a list of training instance indexes that contained missing
   * target values that were replaced via interpolation
   */
  List<Integer> m_missingTargetList;
  
  /**
   *  Holds a list of training instance indexes that contained missing
   * date values (if a date time stamp is being used)
   */
  List<Integer> m_missingTimeStampList;
  
  public String getAlgorithmName() {
    if (m_forecaster != null) {
      String spec = getForecasterSpec();
      spec = spec.replace("weka.classifiers.", "");
      spec = spec.replace("functions.", "");
      spec = spec.replace("bayes.", "");
      spec = spec.replace("rules.", "");
      spec = spec.replace("trees.", "");
      spec = spec.replace("meta.", "");
      spec = spec.replace("lazy.", "");
      spec = spec.replace("supportVector.", "");
      return spec;
    }
    
    return "";
  }
  
  public void setTSLagMaker(TSLagMaker lagMaker) {
    m_lagMaker = lagMaker;
  }
  
  public TSLagMaker getTSLagMaker() {
    return m_lagMaker;
  }
      
  /**
   * Returns an enumeration describing the available options.
   *
   * @return an enumeration of all the available options.
   */
  public Enumeration<Option> listOptions() {
    Vector<Option> newVector = new Vector<Option>();
    
    newVector.add(new Option("\tSet the fields to forecast.", 
                 "F", 1, "-F <comma separated list of names>"));
    
    newVector.add(new Option("\tSet the minimum lag length to generate." +
    		"\n\t(default = 1)", "L", 1, "-L <num>"));
    newVector.add(new Option("\tSet the maximum lag length to generate." +
                "\n\t(default = 12)", "M", 1, "-M <num>"));
    newVector.add(new Option("\tFine tune selection of lags within min and " +
        "max by specifying" +
        " ranges", "R", 1, "-R <ranges>"));
    newVector.add(new Option("\tAverage consecutive long lags.", "A", 0, "-A"));
    newVector.add(new Option("\tAverage those lags longer than this number of" +
    		"time steps.\n\tUse in conjuction with -A.\n\t" +
    		"(default = 2)", "B", 1, "-B <num>"));
    newVector.add(new Option("\tAverage this many consecutive long lags.\n\t" +
    		"Use in conjuction with -B (default = 2)", "C", 1, "-C <num>"));
    newVector.add(new Option("\tDon't adjust for trends.", "Z", 0, "-Z"));
    newVector.add(new Option("\tSpecify the name of the timestamp field", "G", 
        1, "-G <timestamp name>"));
    newVector.add(new Option("\tAdjust for variance.", "V", 0, "-V"));
    newVector.add(new Option("\tSpecify the primary periodic field, " +
    		"\n\tif one exists already in the data " +
    		"(e.g. day, month, quarter etc.\n\tIf there is more than" +
    		"one such field, choose the one with the finest granularity.\n\t" +
    		"This field must be" +
    		"cyclic and declared as nominal.", "periodic", 
    		1, "-periodic <field name>"));
    newVector.add(new Option("\tCalculate confidence limits for predictions\n\t" +
    		"(based on errors) for up to, and including, " +
    		"the specified\n\tnumber of time steps" +
    		"into the future\n\t(default = 0 (don't compute conf. levels)).", 
    		"conf", 1, "-conf <num steps>"));
    newVector.add(new Option("\tConfidence level for computing confidence limits.\n\t" +
    		"Use in conjunction with -conf.\n\t(default = 0.95).", "P", 1,
    		"-P <confidence level>"));
    newVector.add(new Option("\tSpecify the base regression scheme to use.\n\t" +
    		"Supply a fully qualified name, along with options, enclosed in\n\t" +
    		"quotes (e.g. \"weka.classifiers.functions.SMOreg -R 0.5\")." +
    		"\n\t(default = weka.classifiers.functions.SMOreg)",
    		"W", 1, "-W"));
    newVector.add(new Option("\tAdd an AM/PM indicator (requires a date timestamp)",
        "am-pm", 0, "-am-pm"));
    newVector.add(new Option("\tAdd a day of the week field (requres a date" +
                " timestamp)", "day", 0, "-day"));
    newVector.add(new Option("\tAdd a weekend indicator (requires a date timestamp)",
        "weekend", 0, "-weekend"));
    newVector.add(new Option("\tAdd a month field (requires a date timestamp)",
        "month", 0, "-month"));
    newVector.add(new Option("\tAdd a quarter of the year field (" +
                "requires a date timestamp)", "quarter", 0, "-quarter"));
    
    return newVector.elements();
  }
  
  public void setOptions(String[] options) throws Exception {
    
    String fieldsToForecast = Utils.getOption('F', options);
    if (fieldsToForecast.length() == 0) {
      throw new Exception("Must specify the name of at least one field to forecast!");
    }
    
    setFieldsToForecast(fieldsToForecast);
    
    String minL = Utils.getOption('L', options);
    if (minL.length() > 0) {
      int mL = Integer.parseInt(minL);
      //setMinLag(mL);
      m_lagMaker.setMinLag(mL);
      if (mL < 1) {
        throw new Exception("Minimum lag can't be less than 1!");        
      }
    }
    
    String maxL = Utils.getOption('M', options);
    if (maxL.length() > 0) {
      int mL = Integer.parseInt(maxL);
      //setMaxLag(mL);
      m_lagMaker.setMaxLag(mL);
    }
    
    if (m_lagMaker.getMaxLag() < m_lagMaker.getMinLag()) {
      throw new Exception("Can't have the maximum lag set lower than the minimum lag!");
    }
    
    String lagRange = Utils.getOption('R', options);
    if (lagRange.length() > 0) {
      m_lagMaker.setLagRange(lagRange);
    }
    
    boolean avLongLags = Utils.getFlag('A', options);
    //setAverageConsecutiveLongLags(!dontAv);
    m_lagMaker.setAverageConsecutiveLongLags(avLongLags);
    
    String avLongerThan = Utils.getOption('B', options);
    if (avLongerThan.length() > 0) {
      int avL = Integer.parseInt(avLongerThan);
      if (avL < m_lagMaker.getMinLag() || avL > m_lagMaker.getMaxLag()) {
        throw new Exception("Average consecutive long lags value can't " +
        		"be less than the minimum lag or greater than the " +
        		"maximum lag!");
      }
      //setAverageLagsAfter(avL);
      m_lagMaker.setAverageLagsAfter(avL);
    }
    
    String consecutiveLongLagS = Utils.getOption('C', options);
    if (consecutiveLongLagS.length() > 0) {
      int consecutive = Integer.parseInt(consecutiveLongLagS);
      if (consecutive < 1 || 
          consecutive > (m_lagMaker.getMaxLag() - m_lagMaker.getMinLag())) {
        throw new Exception("Number of consecutive long lags to average " +
        		"must be greater than 0 and less than " 
            + (m_lagMaker.getMaxLag() - m_lagMaker.getMinLag()));
      }
      //setNumConsecutiveLongLagsToAverage(consecutive);
      m_lagMaker.setNumConsecutiveLongLagsToAverage(consecutive);
    }
    
    boolean dontAdjTrends = Utils.getFlag('Z', options);
    //setAdjustForTrends(!dontAdjTrends);
    m_lagMaker.setAdjustForTrends(!dontAdjTrends);
    
    boolean adjVariance = Utils.getFlag("V", options);
    //setAdjustForVariance(!dontAdjVariance);
    m_lagMaker.setAdjustForVariance(adjVariance);
    
    String timeStampF = Utils.getOption('G', options);
    if (timeStampF.length() > 0) {
      m_lagMaker.setTimeStampField(timeStampF);
    }
    
    setAddAMIndicator(Utils.getFlag("am-pm", options));
    setAddDayOfWeek(Utils.getFlag("day", options));
    setAddWeekendIndicator(Utils.getFlag("weekend", options));
    setAddMonthOfYear(Utils.getFlag("month", options));
    setAddQuarterOfYear(Utils.getFlag("quarter", options));
    
    String primaryPeriodicN = Utils.getOption("periodic", options);
    if (primaryPeriodicN.length() > 0) {
      setPrimaryPeriodicFieldName(primaryPeriodicN);
    }
    
    String confSteps = Utils.getOption("conf", options);
    if (confSteps.length() > 0) {
      int numSteps = Integer.parseInt(confSteps);
      if (numSteps < 0) {
        throw new Exception("Number of steps must be >= 0");
      }
      setCalculateConfLimitsForForecasts(numSteps);
    }
    
    String confLevel = Utils.getOption('P', options);
    if (confLevel.length() > 0) {
      double cL = Double.parseDouble(confLevel);
      if (cL < 0 || cL > 1) {
        throw new Exception("Confidence level must be between 0 and 1.");
      }
      setConfidenceLevel(cL);      
    }
    
    String baseClassifierS = Utils.getOption('W', options);
    if (baseClassifierS.length() == 0) {
      baseClassifierS = "weka.classifiers.functions.SMOreg";
    }
    String[] classifierSpec = Utils.splitOptions(baseClassifierS);
    if (classifierSpec.length == 0) {
      throw new Exception("Invalid classifier specification.");
    }
    String classifierName = classifierSpec[0];
    classifierSpec[0] = "";
    setBaseForecaster(AbstractClassifier.forName(classifierName, classifierSpec));
  }
  
  /**
   * Gets the current settings of this Forecaster.
   *
   * @return an array of strings suitable for passing to setOptions
   */
  public String[] getOptions() {
    ArrayList<String> options = new ArrayList<String>();
    
    List<String> fieldsToForecast = m_lagMaker.getFieldsToLag();
    options.add("-F"); options.add(fieldsToForecast.toString());
    
    options.add("-L"); options.add("" + m_lagMaker.getMinLag());
    options.add("-M"); options.add("" + m_lagMaker.getMaxLag());
    
    if (m_lagMaker.getLagRange().length() > 0) {
      options.add("-R"); options.add(m_lagMaker.getLagRange());
    }
    
    if (m_lagMaker.getAverageConsecutiveLongLags()) {
      options.add("-A");
    } else {
      options.add("-B"); options.add("" + m_lagMaker.getAverageLagsAfter());
    }
    
    options.add("-C");
    options.add("" + m_lagMaker.getNumConsecutiveLongLagsToAverage());
    
    if (!m_lagMaker.getAdjustForTrends()) {
      options.add("-Z");
    }
    
    if (m_lagMaker.getAdjustForVariance()) {
      options.add("-V");
    }
    
    if (getTimeStampField() != null && getTimeStampField().length() > 0) {
      options.add("-G"); options.add(getTimeStampField());
    }
    
    if (getAddAMIndicator()) {
      options.add("-am-pm");
    }
    
    if (getAddDayOfWeek()) {
      options.add("-day");
    }
    
    if (getAddWeekendIndicator()) {
      options.add("-weekend");
    }
    
    if (getAddMonthOfYear()) {
      options.add("-month");
    }
    
    if (getAddQuarterOfYear()) {
      options.add("-quarter");
    }
    
    if (getPrimaryPeriodicFieldName() != null && 
        getPrimaryPeriodicFieldName().length() > 0) {
      options.add("-periodic"); options.add(getPrimaryPeriodicFieldName());
    }
    
    options.add("-conf");
    options.add("" + getCalculateConfLimitsForForecasts());
    
    options.add("-P");
    options.add("" + getConfidenceLevel());
    
    options.add("-W"); options.add(getForecasterSpec());
    
    return options.toArray(new String[1]);
  }
  
  protected String getForecasterSpec() {
    Classifier c = getBaseForecaster();
    
    if (c instanceof OptionHandler) {
      return c.getClass().getName() + " " 
        + Utils.joinOptions(((OptionHandler)c).getOptions());
    } else {
      return c.getClass().getName();
    }
  }            
  
  public void setFieldsToForecast(String fieldsToForecast) throws Exception {
    super.setFieldsToForecast(fieldsToForecast);
    m_lagMaker.setFieldsToLag(m_fieldsToForecast);
  }  
  
  public void setTimeStampField(String name) {
    m_lagMaker.setTimeStampField(name);
  }
  
  public String getTimeStampField() {
    return m_lagMaker.getTimeStampField();
  }
  
  public void setAddAMIndicator(boolean am) {
    m_lagMaker.setAddAMIndicator(am);
  }
  
  public boolean getAddAMIndicator() {
    return m_lagMaker.getAddAMIndicator();
  }
  
  public void setAddDayOfWeek(boolean d) {
    m_lagMaker.setAddDayOfWeek(d);
  }
  
  public boolean getAddDayOfWeek() {
    return m_lagMaker.getAddDayOfWeek();
  }
  
  public void setAddWeekendIndicator(boolean w) {
    m_lagMaker.setAddWeekendIndicator(w);
  }
  
  public boolean getAddWeekendIndicator() {
    return m_lagMaker.getAddWeekendIndicator();
  }
  
  public void setAddMonthOfYear(boolean m) {
    m_lagMaker.setAddMonthOfYear(m);
  }
  
  public boolean getAddMonthOfYear() {
    return m_lagMaker.getAddMonthOfYear();
  }
  
  public void setAddQuarterOfYear(boolean q) {
    m_lagMaker.setAddQuarterOfYear(q);
  }
  
  public boolean getAddQuarterOfYear() {
    return m_lagMaker.getAddQuarterOfYear();
  }
  
  public void setPrimaryPeriodicFieldName(String p) {
    //m_primaryPeriodicName = p;
    m_lagMaker.setPrimaryPeriodicFieldName(p);
  }
  
  public String getPrimaryPeriodicFieldName() {
    return m_lagMaker.getPrimaryPeriodicFieldName();
  }
  
  public void setCalculateConfLimitsForForecasts(int steps) {
    m_calculateConfLimitsSteps = steps;
  }
  
  public int getCalculateConfLimitsForForecasts() {
    return m_calculateConfLimitsSteps;
  }
  
  public void setConfidenceLevel(double confLevel) {
    m_confidenceLevel = confLevel;
  }
  
  public double getConfidenceLevel() {
    return m_confidenceLevel;
  }
  
  public void setBaseForecaster(Classifier f) {
    m_forecaster = f;
  }
  
  public Classifier getBaseForecaster() {
    return m_forecaster;
  }  

  public void reset() {    
    m_modelBuilt = false;
    
    /*m_varianceAdjusters = null;
    m_lagMakers = null;
    m_averagedLagMakers = null;  
    m_timeIndexMakers = null;
    m_timeLagCrossProductMakers = null;*/
    
    m_lagMaker.reset();
    m_dateRemover = null;
    
    m_primedInput = null;
    m_confidenceLimitEstimator = null;
    m_missingTargetList = new ArrayList<Integer>();
    m_missingTimeStampList = new ArrayList<Integer>();
  }
  
  protected class SingleTargetForecaster implements Serializable {
    
    private static final long serialVersionUID = -4404412501006669036L;

    protected Classifier m_targetForecaster;
    
    private Remove m_otherTargetRemover;
    private int m_classIndex;
    private String m_className;
    
    public void setClassifier(Classifier classifier) {
      m_targetForecaster = classifier;
    }
    
    // assumes transformed data
    public void buildForecaster(Instances train, String targetName, 
        PrintStream... progress) 
      throws Exception {
      
      if (m_targetForecaster == null) {
        throw new Exception("[SingleTargetForecaster] base classifier has" +
        		" not been set!");
      }
      
      train = new Instances(train);
      m_classIndex = train.attribute(targetName).index();
      
      if (m_classIndex < 0) {
        throw new Exception("Can't find target field '" + targetName + "' in" +
        "the data!");        
      }
      
      if (!train.attribute(m_classIndex).isNumeric()) {
        throw new Exception("[SingleTargetForecaster] target '" + targetName
            + "' is not numeric!");
      }
      
      train.setClassIndex(m_classIndex);
      m_className = targetName;
      
      String otherTargets = "";
      for (String n : m_fieldsToForecast) {
        if (!n.equals(targetName)) {
          int i = train.attribute(n).index();
          if (i >= 0) {
            otherTargets += (i + 1) + ",";
          }
        }
      }
      
      if (otherTargets.length() > 0) {
        otherTargets = otherTargets.substring(0, otherTargets.lastIndexOf(','));
        m_otherTargetRemover = new Remove();
        m_otherTargetRemover.setAttributeIndices(otherTargets);
        m_otherTargetRemover.setInputFormat(train);
        train = Filter.useFilter(train, m_otherTargetRemover);
      } 
      
      for (PrintStream p : progress) {
        p.println("Building forecaster for target: " + m_className);
      }
      
      m_targetForecaster.buildClassifier(train);
    }
    
    public double forecastOneStepAhead(Instance transformed) throws Exception {
      transformed.dataset().setClassIndex(m_classIndex);
      
      if (m_otherTargetRemover != null) {
        m_otherTargetRemover.input(transformed);
        transformed = m_otherTargetRemover.output();
      }
      double pred = m_targetForecaster.classifyInstance(transformed);
      
      // undo the log if adjusting for variance
      if (m_lagMaker.getAdjustForVariance()) {
        pred = Math.exp(pred);
      }
      return pred;
    }
    
    public String toString() {
      if (m_targetForecaster == null) {
        return "SingleTargetForecaster: no model built yet!";
      }
      
      return m_className + ":\n" + m_targetForecaster.toString();
    }
  }
  
  public void buildForecaster(Instances insts, PrintStream... progress) 
    throws Exception {

    reset();
    m_originalHeader = new Instances(insts, 0);
    insts = m_lagMaker.replaceMissing(insts, false, 
        m_missingTargetList, m_missingTimeStampList);
    
/*    int classIndex = insts.attribute(m_fieldsToForecast.get(0)).index();
    if (classIndex < 0) {
      throw new Exception("Can't find target field '" + m_fieldsToForecast + "' in" +
      		"the data!");
    } */
    
    //setupPeriodicMaps(insts);
    
    for (PrintStream p : progress){
      p.println("Transforming input data...");
    }
    
    //Instances trainingData = removeExtraneousAttributes(insts);
    Instances trainingData = insts;
    trainingData = m_lagMaker.getTransformedData(trainingData);
    m_transformedHeader = new Instances(trainingData, 0);
    
    m_dateRemover = new RemoveType();
    m_dateRemover.setOptions(new String[] {"-T", "date"});
    m_dateRemover.setInputFormat(trainingData);
    trainingData = Filter.useFilter(trainingData, m_dateRemover);
    //System.err.println(trainingData);
    
//    m_lastHistoricInstance = insts.instance(insts.numInstances() - 1);
    
    m_singleTargetForecasters = new ArrayList<SingleTargetForecaster>();
    for (int i = 0; i < m_fieldsToForecast.size(); i++) {
      SingleTargetForecaster f = new SingleTargetForecaster();
      Classifier c = AbstractClassifier.makeCopy(m_forecaster);
      f.setClassifier(c);
      f.buildForecaster(trainingData, m_fieldsToForecast.get(i));
      m_singleTargetForecasters.add(f);
    }
    
/*    classIndex = trainingData.attribute(m_fieldsToForecast.get(0)).index();
    trainingData.setClassIndex(classIndex);
    m_forecaster.buildClassifier(trainingData); */
    
    m_modelBuilt = true;
    /*for (int i = 0; i < m_singleTargetForecasters.size(); i++) {
      System.out.println(m_singleTargetForecasters.get(i));
    }*/
    

    if (m_calculateConfLimitsSteps > 0) {
      for (PrintStream p : progress){
        p.println("Computing confidence intervals...");
      }
      // -1 indicates not using an artificial time index
      int artificialTimeStart = (m_lagMaker.isUsingAnArtificialTimeIndex())
        ? 1 : -1;
      ErrorBasedConfidenceLimitEstimator e = new ErrorBasedConfidenceLimitEstimator();
      e.calculateConfidenceOffsets(this, insts, 
          m_lagMaker.getMaxLag(), artificialTimeStart, 
          m_calculateConfLimitsSteps, m_confidenceLevel, progress);
      m_confidenceLimitEstimator = e;
    }    
    //System.out.println(trainingData);
  }
  
  public String toString() {
    if (!m_modelBuilt) {
      return "Forecaster has not been built yet!";
    }
    
    StringBuffer result = new StringBuffer();
    result.append("Transformed training data:\n\n");
    for (int i = 0; i < m_transformedHeader.numAttributes(); i++) {
      result.append("              " 
          + m_transformedHeader.attribute(i).name()).append("\n");
    }
    
    if (m_missingTargetList != null && m_missingTargetList.size() > 0) {
      Collections.sort(m_missingTargetList);
      result.append("\n---------------------------------------------------\n" +
      		"The following training instances had missing values\n" +
      		"imputed via interpolation. Check source data as\n" +
      		"this may affect forecasting performance:\n\n");
      for (int i = 0; i < m_missingTargetList.size(); i++) {
        if (i == 0) {
          result.append("              " + m_missingTargetList.get(i));
        } else if (!m_missingTargetList.get(i).equals(m_missingTargetList.get(i - 1))) {
          result.append("," + m_missingTargetList.get(i));
        }
      }
      result.append("\n---------------------------------------------------\n");
    }
    
    if (m_missingTimeStampList != null && m_missingTimeStampList.size() > 0) {
      Collections.sort(m_missingTimeStampList);
      result.append("\n--------------------------------------------------------\n" +
      		"The following training instances had missing time stamps:\n\n");
      for (int i = 0; i < m_missingTimeStampList.size(); i++) {
        if (i == 0) {
          result.append("              " + m_missingTimeStampList.get(i));
        } else {
          result.append("," + m_missingTimeStampList.get(i));
        }
      }
      result.append("\n-------------------------------------------------------\n");
    }
    
    // System.out.println(m_transformedHeader);
    for (int i = 0; i < m_singleTargetForecasters.size(); i++) {
      result.append("\n"
          + m_singleTargetForecasters.get(i)).append("\n");
    }
    
    return result.toString();
  }
  
  protected Instance applyFilters(Instance source, boolean incrementArtificialTime,
      boolean setAnyPeriodic) throws Exception {
    Instance result = source;
    
    /*if (m_extraneousAttributeRemover != null) {
      m_extraneousAttributeRemover.input(result);
      result = m_extraneousAttributeRemover.output();
    }  */ 
    
    result = m_lagMaker.processInstance(result, incrementArtificialTime, setAnyPeriodic);
    
    return result;
  }

  @Override
  public void primeForecaster(Instances insts) throws Exception {

    m_primedInput = new Instances(insts);
    
//    m_lastHistoricInstance = m_primedInput.instance(m_primedInput.numInstances() - 1);    
    m_lagMaker.clearLagHistories();
    
    for (int i = 0; i < m_primedInput.numInstances(); i++) {
      applyFilters(m_primedInput.instance(i), false, false);
    }
  }
  
  public void primeForecasterIncremental(Instance inst) throws Exception {
    if (m_primedInput == null) {
      throw new Exception("WekaForecaster hasn't been initialized with "
          + "a call to primeForecaster()!!");
    }
    
    applyFilters(inst, false, false);
  }
  
  protected double forecastOneStepAhead(Instance transformed) throws Exception {
    return m_forecaster.classifyInstance(transformed);    
  }

  @Override
  public List<List<NumericPrediction>> forecast(int numSteps, 
      PrintStream... progress) throws Exception {
    
    // we need to:
    // 1) input a new instance with ? for target into the filter chain in order to push the most recent
    // known target value into the history
    // 2) output() from filter
    // 3) make the t + 1 prediction
    // 4) set the value of the target for the input instance (this instance, now stored in
    // the history buffer of the TimeseriesTranslate filters, will now have the predicted target
    // value - hopefully)
    
    // 4 won't work. Need to add the input instance (with prediction set) to the end
    // of the primed input data set and then call primeForecaster() again
    
    //double[] finalForecast = new double[numSteps];
    List<List<NumericPrediction>> forecastForSteps = new ArrayList<List<NumericPrediction>>();
    
    for (int i = 0; i < numSteps; i++) {
      // set the target to missing first
      double[] newVals = new double[m_originalHeader.numAttributes()];
      // set all to missing
      for (int j = 0; j < newVals.length; j++) {
        newVals[j] = Utils.missingValue();
      }

      // create the test instance (original format)
      Instance origTest = new DenseInstance(1.0, newVals);
      origTest.setDataset(m_originalHeader);
      
//      System.err.println("Original with periodic set " + origTest);

      Instance transformedWithDate = origTest;

      // do all the filters
//      System.err.println("--- " + transformed);
      
      
      //transformedWithDate = applyFilters(transformedWithDate, true, true);
      transformedWithDate = m_lagMaker.processInstancePreview(transformedWithDate, true, true);
      
      // the data timestamp (if exists) has now been remapped, so we can remove
      // the original
      m_dateRemover.input(transformedWithDate);
      Instance transformed = m_dateRemover.output();
      
      //System.err.println(transformedWithDate.dataset());
      //System.err.println(transformedWithDate);
      
//      System.err.println("Transformed: " + transformed);

      // get a prediction
      double[] preds = new double[m_singleTargetForecasters.size()];
      for (int j = 0; j < m_singleTargetForecasters.size(); j++) {
        preds[j] = m_singleTargetForecasters.get(j).forecastOneStepAhead(transformed);
      }
      
//      double pred = forecastOneStepAhead(transformed);
/*      // undo the log if adjusting for variance
      if (m_lagMaker.getAdjustForVariance()) {
        pred = Math.exp(pred);
      } */
      //finalForecast[i] = pred;
      
      // predictions at step i for all the targets (can only handle a single target at
      // present)
      List<NumericPrediction> finalForecast = new ArrayList<NumericPrediction>();
      
      // add confidence limits (if applicable)
      for (int j = 0; j < m_fieldsToForecast.size(); j++) {        
        if (m_confidenceLimitEstimator != null && i < m_calculateConfLimitsSteps) {
          double[] limits =
            m_confidenceLimitEstimator.getConfidenceLimitsForTarget(m_fieldsToForecast.get(j), 
                preds[j], i + 1);
          double[][] limitsToAdd = new double[1][];
          limitsToAdd[0] = limits;
          finalForecast.add(new NumericPrediction(Utils.missingValue(), preds[j], 1.0, limitsToAdd));
        } else {
          finalForecast.add(new NumericPrediction(Utils.missingValue(), preds[j]));
        }
      }
      forecastForSteps.add(finalForecast);

      // set the value of the target in the original test instance
      for (int j = 0; j < m_fieldsToForecast.size(); j++) {
        int targetIndex = m_originalHeader.attribute(m_fieldsToForecast.get(j)).index();
        origTest.setValue(targetIndex, preds[j]);
      }
      
      // If we have a real time stamp, then set the incremented value in the original
      // test instance
      if (!m_lagMaker.isUsingAnArtificialTimeIndex() &&
          m_lagMaker.getAdjustForTrends() &&
          m_lagMaker.getTimeStampField() != null 
          && m_lagMaker.getTimeStampField().length() > 0) {
        int timeIndex = m_originalHeader.attribute(m_lagMaker.getTimeStampField()).index();
        double timeValue = 
          transformedWithDate.value(transformedWithDate.dataset().attribute(m_lagMaker.getTimeStampField()));
        origTest.setValue(timeIndex, timeValue);
      }

      // add it to the end of the primed input data
      m_primedInput.delete(0);
      m_primedInput.add(origTest);
      m_primedInput.compactify();

      // now re-prime the forecaster
      //primeForecaster(m_primedInput);
      primeForecasterIncremental(origTest);
    }
    
    //TODO fix this - move to eval class?
    if (m_lagMaker.isUsingAnArtificialTimeIndex()) {
      m_lagMaker.incrementArtificialTimeValue(-(numSteps - 1));
        // -= (numSteps - 1);
    }
    
    return forecastForSteps;
  }
  
  public static void main(String[] args) {
    try {
      /*Instances train = new Instances(new BufferedReader(new FileReader(args[0])));
      WekaForecaster wf = new WekaForecaster();
      ArrayList<String> fieldsToForecast = new ArrayList<String>();
      fieldsToForecast.add(args[1]);
      wf.setFieldsToForecast(fieldsToForecast);
      wf.setPrimaryPeriodicFieldName(args[2]);
      Instances trans = wf.getTransformedData(train);
      System.out.println(trans); */
      
      WekaForecaster fs = new WekaForecaster();
      fs.runForecaster(fs, args);
    } catch (Exception ex) {
      ex.printStackTrace();
    }
  }
}
