package weka.classifiers.timeseries.core;

import java.io.Serializable;
import java.util.Calendar;
import java.util.Date;
import java.util.GregorianCalendar;

public class CustomPeriodicTest implements Serializable {
  /**
   * For serialization
   */
  private static final long serialVersionUID = -1470297629040276248L;

  protected String m_lowerBoundTest;
  
  protected String m_upperBoundTest;
  
  enum Operator {
    NONE("None") {
      boolean evaluate(int first, int second) {
        return false;
      }
    },
    EQUALS("=") {
      boolean evaluate(int first, int second) {
        return (first == second);
      }
    },
    GREATER_THAN_OR_EQUAL_TO(">=") {
      boolean evaluate(int first, int second) {
        return (first >= second);
      }
    },
    GREATER_THAN(">") {
      boolean evaluate(int first, int second) {
        return (first > second);
      }
    },
    LESS_THAN_OR_EQUAL_TO("<=") {
      boolean evaluate(int first, int second) {
        return (first <= second);
      }
    },
    LESS_THAN("<") {
      boolean evaluate(int first, int second) {
        return (first < second);
      }
    };
    
    private final String m_stringVal;
    
    abstract boolean evaluate(int first, int second);
    
    Operator(String name) {
      m_stringVal = name;
    }
    
    public String toString() {
      return m_stringVal;
    }
  }
  
  protected class TestPart implements Serializable {
    /**
     * For serialization
     */
    private static final long serialVersionUID = -898852333853148631L;

    protected Operator m_boundOperator = Operator.NONE;
    
    protected int m_year = -100;
    protected int m_month = -100;
    protected int m_week_of_yr = -100;
    protected int m_week_of_month = -100;
    protected int m_day_of_yr = -100;
    protected int m_day_of_month = -100;
    protected int m_day_of_week = -100;
    protected int m_hour_of_day = -100;
    protected int m_min_of_hour = -100;
    protected int m_second = -100;
    
    protected Calendar c = new GregorianCalendar();
    
    // TODO eval(Date d) method
    public boolean eval(Date d) {
      
      c.setTime(d);
      
      boolean result = true;
      
      if (m_year != -100) {
        result = (result && m_boundOperator.
            evaluate(c.get(Calendar.YEAR), m_year));
      }
      if (m_month != -100) {        
        // adjust to make it 1-based rather than 0-based
        result = (result && m_boundOperator.
            evaluate(c.get(Calendar.MONTH) + 1, m_month));
      }
      if (m_week_of_yr != -100) {
        result = (result && m_boundOperator.
            evaluate(c.get(Calendar.WEEK_OF_YEAR), m_week_of_yr));
      }
      if (m_week_of_month != -100) {
        result = (result && m_boundOperator.
            evaluate(c.get(Calendar.WEEK_OF_MONTH), m_week_of_month));
      }
      if (m_day_of_yr != -100) {
        result = (result && m_boundOperator.
            evaluate(c.get(Calendar.DAY_OF_YEAR), m_day_of_yr));
      }
      if (m_day_of_month != -100) {
        result = (result && m_boundOperator.
            evaluate(c.get(Calendar.DAY_OF_MONTH), m_day_of_month));
      }
      if (m_day_of_week != -100) {
        result = (result && m_boundOperator.
            evaluate(c.get(Calendar.DAY_OF_WEEK), m_day_of_week));
      }
      if (m_hour_of_day != -100) {
        result = (result && m_boundOperator.
            evaluate(c.get(Calendar.HOUR_OF_DAY), m_hour_of_day));
      }
      if (m_min_of_hour != -100) {
        result = (result && m_boundOperator.
            evaluate(c.get(Calendar.MINUTE), m_min_of_hour));
      }
      if (m_second != -100) {
        result = (result && m_boundOperator.
            evaluate(c.get(Calendar.SECOND), m_second));
      }
            
      return result;
    }
  }
  
  protected TestPart m_upperTest;
  
  protected TestPart m_lowerTest;
  
  public CustomPeriodicTest(String theTest) throws IllegalArgumentException {
    setTest(theTest);
  }
  
  public void setTest(String theTest) throws IllegalArgumentException {
    // check for brackets first
    if (theTest.indexOf('(') < 0 || theTest.indexOf(')') < 0) {
      throw new IllegalArgumentException("Custom periodic test needs enclosing " +
      		"parenthesis!");
    }
      
    // split off brackets
    theTest = theTest.substring(1, theTest.indexOf(')') - 1);    
    
    if (theTest.indexOf(' ') > 0) {
      String[] parts = theTest.split(" ");
      
      m_lowerTest = parseTest(parts[0]);
      m_upperTest = parseTest(parts[1]);
      
      if (m_upperTest.m_boundOperator == m_lowerTest.m_boundOperator) {
        throw new IllegalArgumentException("Doesn't make sense to define " +
        		"an upper and lower test that uses the same operator");
      }      
    } else {
      m_lowerTest = parseTest(theTest);
    }
  }
  
  public boolean evaluate(Date d) {
    boolean result = m_lowerTest.eval(d);
    if (m_upperTest != null) {
      result = (result && m_upperTest.eval(d));
    }
    
    return result;
  }
  
  private int parseMonth(String month) throws IllegalArgumentException {
    // try parsing as a number
    try {
      int result = Integer.parseInt(month);
      return result;
    } catch (NumberFormatException e) {}
    
    // try parsing as a word
    month = month.toLowerCase();
    if (month.equals("january") || month.equals("jan")) {
      return Calendar.JANUARY;
    }
    if (month.equals("february") || month.equals("feb")) {
      return Calendar.FEBRUARY;
    }
    if (month.equals("march") || month.equals("mar")) {
      return Calendar.MARCH;
    }
    if (month.equals("april") || month.equals("apr")) {
      return Calendar.APRIL;
    }
    if (month.equals("may")) {
      return Calendar.MAY;
    }
    if (month.equals("june") || month.equals("jun")) {
      return Calendar.JUNE;
    }
    if (month.equals("july") || month.equals("jul")) {
      return Calendar.JULY;
    }
    if (month.equals("august") || month.equals("aug")) {
      return Calendar.AUGUST;
    }
    if (month.equals("september") || month.equals("sept")) {
      return Calendar.SEPTEMBER;
    }
    if (month.equals("october") || month.equals("oct")) {
      return Calendar.OCTOBER;
    }
    if (month.equals("november") || month.equals("nov")) {
      return Calendar.NOVEMBER;
    }
    if (month.equals("december") || month.equals("dec")) {
      return Calendar.DECEMBER;
    }
    
    throw new IllegalArgumentException("Can't parse month!");
  }
  
  protected TestPart parseTest(String aTest) throws IllegalArgumentException {
    aTest = aTest.trim();
    TestPart newTest = new TestPart();
    
    if (aTest.charAt(0) == '=') {
      newTest.m_boundOperator = Operator.EQUALS;
      aTest = aTest.substring(1, aTest.length());
    } else {
      if (aTest.charAt(0) == '>') {
        newTest.m_boundOperator = Operator.GREATER_THAN;
        aTest = aTest.substring(1, aTest.length());
        if (aTest.charAt(0) == '=') {
          newTest.m_boundOperator = Operator.GREATER_THAN_OR_EQUAL_TO;
          aTest = aTest.substring(1, aTest.length());
        }
      } else if (aTest.charAt(0) == '<') {
        newTest.m_boundOperator = Operator.LESS_THAN;
        aTest = aTest.substring(1, aTest.length());
        if (aTest.charAt(0) == '=') {
          newTest.m_boundOperator = Operator.LESS_THAN_OR_EQUAL_TO;
          aTest = aTest.substring(1, aTest.length());
        }
      }
    }
    
    String[] parts = aTest.split(":");
    // do we have the right number of parts?
    if (parts.length != 10) {
      throw new IllegalArgumentException("Test does not contain 10 parts!");
    }
    if (parts[0].trim().length() > 0) {
      newTest.m_year = Integer.parseInt(parts[0].trim());
    }
    if (parts[1].trim().length() > 0) {
      newTest.m_month = parseMonth(parts[1].trim());
    }
    if (parts[2].trim().length() > 0) {
      newTest.m_week_of_yr = Integer.parseInt(parts[2].trim());
    }
    if (parts[3].trim().length() > 0) {
      newTest.m_week_of_month = Integer.parseInt(parts[3].trim());
    }
    if (parts[4].trim().length() > 0) {
      newTest.m_day_of_yr = Integer.parseInt(parts[4].trim());
    }
    if (parts[5].trim().length() > 0) {
      newTest.m_day_of_month = Integer.parseInt(parts[5].trim());
    }
    if (parts[6].trim().length() > 0) {
      newTest.m_day_of_week = Integer.parseInt(parts[6].trim());
    }
    if (parts[7].trim().length() > 0) {
      newTest.m_hour_of_day = Integer.parseInt(parts[7].trim());
    }
    if (parts[8].trim().length() > 0) {
      newTest.m_min_of_hour = Integer.parseInt(parts[8].trim());
    }
    if (parts[9].trim().length() > 0) {
      newTest.m_second = Integer.parseInt(parts[9].trim());
    }
    
    return newTest;
  }
}
