/**
 * 
 */
package weka.classifiers.timeseries.eval;

import java.util.ArrayList;
import java.util.Collections;
import java.util.List;

import weka.classifiers.evaluation.NumericPrediction;
import weka.core.Instance;
import weka.core.Utils;

/**
 * @author mhall
 *
 */
public class ErrorModule extends TSEvalModule {
  
  /** The predictions for each target */
  protected List<List<NumericPrediction>> m_predictions;
  
  /** The counts of each valid target prediction */
  protected double[] m_counts;

  /* (non-Javadoc)
   * @see weka.classifiers.timeseries.eval.TSEvalModule#reset()
   */
  @Override
  public void reset() {
    if (m_targetFieldNames != null) {
      m_predictions = new ArrayList<List<NumericPrediction>>();
      m_counts = new double[m_targetFieldNames.size()];
      
      for (int i = 0; i < m_targetFieldNames.size(); i++) {
        ArrayList<NumericPrediction> predsForTarget = 
          new ArrayList<NumericPrediction>();
        m_predictions.add(predsForTarget);
      }
    }
  }

  /* (non-Javadoc)
   * @see weka.classifiers.timeseries.eval.TSEvalModule#getEvalName()
   */
  @Override
  public String getEvalName() {
    return "Error";
  }

  /* (non-Javadoc)
   * @see weka.classifiers.timeseries.eval.TSEvalModule#getDescription()
   */
  @Override
  public String getDescription() {
    return "Sum of errors";
  }

  /* (non-Javadoc)
   * @see weka.classifiers.timeseries.eval.TSEvalModule#getDefinition()
   */
  @Override
  public String getDefinition() {
    return "sum(predicted - actual)";
  }
  
  public String toString() {
    return getDescription() + " (" + getEvalName() + ")";
  }

  /* (non-Javadoc)
   * @see weka.classifiers.timeseries.eval.TSEvalModule#evaluateForInstance(double[], weka.core.Instance)
   */
  @Override
  public void evaluateForInstance(List<NumericPrediction> forecasts, Instance inst) 
    throws Exception {
    if (m_predictions == null) {
      throw new Exception("Target fields haven't been set yet!");
    }
    
    if (forecasts.size() != m_targetFieldNames.size()) {
      throw new Exception("The number of forecasted values does not match the" +
      		" number of target fields!");
    }
    
    for (int i = 0; i < m_targetFieldNames.size(); i++) {
      double actualValue = getTargetValue(m_targetFieldNames.get(i), inst);
      double predictedValue = forecasts.get(i).predicted();
      //System.err.println("Actual: " + actualValue + " Predicted: " + predictedValue);
      double[][] intervals = forecasts.get(i).predictionIntervals();
      
      NumericPrediction pred = new NumericPrediction(actualValue, predictedValue, 1, intervals);
      m_predictions.get(i).add(pred);
      
      if (!Utils.isMissingValue(predictedValue) && 
          !Utils.isMissingValue(actualValue)) {
        m_counts[i]++;
      }
    }
  }
  
  public double[] calculateMeasure() throws Exception {
    if (m_predictions == null || m_predictions.get(0).size() == 0) {
      throw new Exception("No predictions have been seen yet!");
    }
    
    double[] result = new double[m_targetFieldNames.size()];
    for (int i = 0; i < m_targetFieldNames.size(); i++) {
      List<NumericPrediction> preds = m_predictions.get(i);

      double sumOfE = 0;
      for (NumericPrediction p : preds) {
        if (!Utils.isMissingValue(p.error())) {
          sumOfE += p.error();
        }
      }
      
      result[i] = sumOfE;
    }
    
    return result;
  }
  
  public double[] countsForTargets() throws Exception {
    if (m_predictions == null || m_predictions.get(0).size() == 0) {
      throw new Exception("No predictions have been seen yet!");
    }
    
    return m_counts;
  }
  
  public List<Double> getErrorsForTarget(String targetName) 
    throws IllegalArgumentException {

    for (int i = 0; i < m_targetFieldNames.size(); i++) {
      if (m_targetFieldNames.get(i).equals(targetName)) {
        ArrayList<Double> errors = new ArrayList<Double>();
        List<NumericPrediction> preds = m_predictions.get(i);
        for (int j = 0; j < preds.size(); j++) {
          Double err = new Double(preds.get(j).error());
          errors.add(err);
        }
        return errors;
      }
    }
    
    throw new IllegalArgumentException("Unknown target: " + targetName);
  }
  
  public List<NumericPrediction> getPredictionsForTarget(String targetName) 
    throws IllegalArgumentException {
    
    for (int i = 0; i < m_targetFieldNames.size(); i++) {
      if (m_targetFieldNames.get(i).equals(targetName)) {
        return m_predictions.get(i);
      }
    }
    
    throw new IllegalArgumentException("Unknown target: " + targetName);
  }
  
  public List<List<NumericPrediction>> getPredictionsForAllTargets() {
    return m_predictions;
  }
  
  public String toSummaryString() throws Exception {
    StringBuffer result = new StringBuffer();
    
    double[] measures = calculateMeasure();
    for (int i = 0; i < m_targetFieldNames.size(); i++) {
      result.append(getDescription() + " (" + m_targetFieldNames.get(i) + "): " 
          + Utils.doubleToString(measures[i], 4) + " (n = " + m_counts[i] + ")");
      result.append("\n");
    }
    
    return result.toString();
  }
}
